from dataclasses import dataclass
from typing import Literal

import jax.numpy as jnp
from jaxtyping import Scalar

from src.dataset.types import BatchedTarget
from src.loss.store import register_loss_function
from .types import AbstractLossFunction, BaseLossConfig


@dataclass
class GatedRegressionConfig(BaseLossConfig):
    name: Literal["gated_regression"]
    weights: tuple[float, float] = (1e-4, 1e-4)


@register_loss_function(cfg=GatedRegressionConfig)
class GatedRegression(AbstractLossFunction):
    @staticmethod
    def __call__(
        cfg: GatedRegressionConfig,
        true_y: BatchedTarget,
        pred_y: BatchedTarget,
    ) -> Scalar:
        """
        Apply mse to the distance to the signal if the true label is a signal
        else loss = exp(-w[1] * pred_y)
        """

        # jax.debug.print("true_dist: {}", true_y.flatten())
        # jax.debug.print("pred_dist: {}", pred_y.flatten())

        is_signal = true_y > 0
        # jax.debug.print("is_signal: {}", is_signal)
        target = jnp.abs(true_y)
        # jax.debug.print("target: {}", target)

        # punish examples that are far from the signal less
        factor = jnp.where(is_signal, jnp.exp(-cfg.weights[0] * target), 0)
        # jax.debug.print("factor: {}", factor)
        # jax.debug.print("l1.1: {}", jnp.square(pred_y - target) * factor)

        l1 = (
            cfg.weights[0]
            * jnp.sum(jnp.square(pred_y - target) * factor)
            / jnp.sum(is_signal)
        )

        # l2 = (
        #     1
        #     / cfg.weights[1]
        #     * jnp.sum(~is_signal * jnp.exp(-cfg.weights[1] * pred_y))
        #     / jnp.sum(~is_signal)
        # )
        # jax.debug.print("l2.1: {}", ~is_signal * jnp.exp(-cfg.weights[1] * target))
        # jax.debug.print("l1: {} l2: {}", l1, l2)

        return l1  # + l2



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Callable

from jaxtyping import Scalar

from src.dataset.types import BatchedTarget

LossFunction = Callable[[BatchedTarget, BatchedTarget], Scalar]


@dataclass
class BaseLossConfig:
    name: str
    weight: float = field(default=1.0, init=False)


class AbstractLossFunction:
    """
    Abstract base class for all losses.

    This class is used to define the interface for all losses.
    """

    @abstractmethod
    def __call__(
        cfg: BaseLossConfig,
        true_y: BatchedTarget,
        pred_y: BatchedTarget,
    ) -> Scalar:
        pass



from typing import Tuple


from src.misc.stores import import_all_siblings
from src.models.sequence_mixer.types import SequenceMixer

from .store import get_sequence_mixer_cfgs_type, RegisteredSequenceMixers

import_all_siblings(__name__, __file__)
SequenceMixerConfig = get_sequence_mixer_cfgs_type()


def get_sequence_mixer_cls(
    cfg: SequenceMixerConfig,
    # input_dim: int,
    # key: PRNGKeyArray,
    **kwargs,
) -> Tuple[SequenceMixer, dict[str, bool]]:
    sequence_mixer = RegisteredSequenceMixers[cfg.name][0]

    # TODO: Philipp removed this because he is a noob. Should probably be restored at some point
    # filter_spec = jax.tree_util.tree_map(
    #    sequence_mixer.filter_spec_lambda(), sequence_mixer
    # )

    return sequence_mixer, None



from dataclasses import dataclass
from typing import List, Literal

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import PRNGKeyArray

from src.models.sequence_mixer import SequenceMixerConfig
from src.models.sequence_mixer.types import SequenceMixer

from .types import Encoder, EncoderBaseConfig
from .store import register_encoder


class GLU(eqx.Module):
    w1: eqx.nn.Linear
    w2: eqx.nn.Linear

    def __init__(self, input_dim, output_dim, key):
        w1_key, w2_key = jr.split(key, 2)
        self.w1 = eqx.nn.Linear(input_dim, output_dim, use_bias=True, key=w1_key)
        self.w2 = eqx.nn.Linear(input_dim, output_dim, use_bias=True, key=w2_key)

    def __call__(self, x):
        return self.w1(x) * jax.nn.sigmoid(self.w2(x))


class LinossEncoderBlock(eqx.Module):
    norm: eqx.nn.BatchNorm
    sequence_mixer: SequenceMixer
    glu: GLU
    drop: eqx.nn.Dropout

    def __init__(
        self,
        H,
        drop_rate,
        *,
        input_dtype: jnp.dtype,
        sequence_mixer: SequenceMixer,
        key,
    ):
        sequence_mixerkey, glukey = jr.split(key, 2)
        # self.norm = eqx.nn.BatchNorm(
        #     input_size=H,
        #     axis_name="batch",
        #     channelwise_affine=False,
        #     mode="ema",
        #     dtype=input_dtype,
        # )
        self.norm = eqx.nn.LayerNorm(H)
        self.sequence_mixer = sequence_mixer
        self.glu = GLU(H, H, key=glukey)
        self.drop = eqx.nn.Dropout(p=drop_rate)

    def __call__(self, x, state, *, key):
        """Compute LinOSS block."""
        dropkey1, dropkey2 = jr.split(key, 2)
        skip = x
        x = self.sequence_mixer(x)
        x, state = jax.vmap(self.norm)(x, state)
        # x = x.T
        x = self.drop(jax.nn.gelu(x), key=dropkey1)  # 12*12=144
        x = jax.vmap(self.glu)(x)
        x = self.drop(x, key=dropkey2)
        x = skip + x

        return x, state


@dataclass
class LinossEncoderConfig(EncoderBaseConfig):
    name: Literal["linoss_encoder"]
    num_blocks: int
    hidden_dim: int
    sequence_mixer: SequenceMixerConfig
    use_embedding: bool = False
    pass_through: bool = False
    dropout_rate: float = 0.05


@register_encoder(cfg=LinossEncoderConfig)
class LinossEncoder(Encoder):
    linear_encoder: eqx.nn.Linear
    blocks: List[LinossEncoderBlock]
    embedding: eqx.nn.Embedding | None
    hidden_dim: int
    use_embedding: bool
    pass_through: bool

    def __init__(
        self,
        input_dim: int,
        cfg: LinossEncoderConfig,
        input_dtype: jnp.dtype,
        key: PRNGKeyArray,
        sequence_mixer_cls: SequenceMixer,
        num_embeddings: int | None = None,
        **kwargs,
    ):
        self.hidden_dim = cfg.hidden_dim
        self.use_embedding = cfg.use_embedding
        self.pass_through = cfg.pass_through

        key, embedding_key = jr.split(key, 2)
        if cfg.use_embedding:
            self.embedding = eqx.nn.Embedding(
                num_embeddings=num_embeddings,
                embedding_size=cfg.hidden_dim,
                key=embedding_key,
            )
        else:
            self.embedding = None

        key, linear_encoder_key, *block_keys = jr.split(key, cfg.num_blocks + 2)
        key, *mixer_keys = jr.split(key, cfg.num_blocks + 1)
        self.linear_encoder = eqx.nn.Linear(
            input_dim, cfg.hidden_dim, key=linear_encoder_key, dtype=input_dtype
        )
        self.blocks = [
            LinossEncoderBlock(
                cfg.hidden_dim,
                input_dtype=input_dtype,
                key=b_key,
                sequence_mixer=sequence_mixer_cls(
                    cfg=cfg.sequence_mixer,
                    input_dim=cfg.hidden_dim,
                    input_dtype=input_dtype,
                    key=m_key,
                    **kwargs,
                ),
                drop_rate=cfg.dropout_rate,
            )
            for (m_key, b_key) in zip(mixer_keys, block_keys)
        ]

    def __call__(self, x, state, key):
        """ """
        dropout_keys = jr.split(key, len(self.blocks))
        if self.use_embedding:
            y = jax.vmap(self.embedding)(x[:, 0].astype(jnp.int32))
        else:
            y = jax.vmap(self.linear_encoder)(x)  # 6

        intermediate_results = [y]
        for i, (block, d_key) in enumerate(zip(self.blocks, dropout_keys)):
            # x shape is (timesteps, hidden_dim)
            y, state = block(y, state, key=d_key)
            intermediate_results.append(y)

        if self.pass_through:
            y = jnp.concatenate(intermediate_results, axis=1)

        return y, state

    @property
    def output_dim(self) -> int:
        if self.pass_through:
            return self.hidden_dim * (len(self.blocks) + 1)
        else:
            return self.hidden_dim



from abc import abstractmethod
from dataclasses import dataclass
from typing import Callable

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from src.dataset.types import EncoderOutput, UnbatchedInput
from src.models.sequence_mixer.types import SequenceMixer


@dataclass
class EncoderBaseConfig:
    name: str


class Encoder(eqx.Module):
    """
    Abstract base class for all encoders.

    This class is used to define the interface for all encoders.
    """

    @abstractmethod
    def __init__(
        self,
        input_dim: int,
        cfg: EncoderBaseConfig,
        input_dtype: jnp.dtype,
        key: PRNGKeyArray,
        sequence_mixer: SequenceMixer | None = None,
        num_embeddings: int | None = None,
        **kwargs,
    ) -> None:
        pass

    @abstractmethod
    def __call__(
        self, x: UnbatchedInput, state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[EncoderOutput, eqx.nn.State]:
        pass

    def filter_spec_lambda(self) -> Callable[..., bool]:
        return lambda _: True

    @property
    @abstractmethod
    def output_dim(self) -> int:
        pass



from typing import Tuple

import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from src.misc.stores import import_all_siblings
from src.models.encoder.types import Encoder
from src.models.sequence_mixer.types import SequenceMixer

from src.models.encoder.store import RegisteredEncoders, get_encoder_cfgs_type

import_all_siblings(__name__, __file__)
EncoderConfig = get_encoder_cfgs_type()


def get_encoder(
    cfg: EncoderConfig,
    key: PRNGKeyArray,
    input_dim: int,
    input_dtype: jnp.dtype,
    sequence_mixer_cls: SequenceMixer | None = None,
    num_embeddings: int | None = None,
    **kwargs,
) -> Tuple[Encoder, dict[str, bool]]:
    encoder_cls, _ = RegisteredEncoders[cfg.name]
    encoder = encoder_cls(
        input_dim=input_dim,
        cfg=cfg,
        input_dtype=input_dtype,
        key=key,
        sequence_mixer_cls=sequence_mixer_cls,
        num_embeddings=num_embeddings,
        **kwargs,
    )
    filter_spec = jax.tree_util.tree_map(encoder.filter_spec_lambda(), encoder)
    return encoder, filter_spec



from typing import List

import jax.random as jr
from jaxtyping import PRNGKeyArray

from .types import Head, HeadBaseConfig

from src.misc.stores import import_all_siblings

from .store import RegisteredHeads, get_head_cfgs_type

import_all_siblings(__name__, __file__)
HeadConfig = get_head_cfgs_type()


def get_heads(
    *cfgs: HeadBaseConfig,
    key: PRNGKeyArray,
    input_dim: int,
    output_dim: int,
    **kwargs,
) -> List[Head]:
    keys = jr.split(key, len(cfgs))
    return [
        RegisteredHeads[cfg.name][0](
            input_dim=input_dim,
            output_dim=cfg.output_dim if cfg.output_dim is not None else output_dim,
            cfg=cfg,
            key=key,
            **kwargs,
        )
        for cfg, key in zip(cfgs, keys)
    ]



from typing import Type
import functools as ft

from src.misc.stores import get_cfgs_type, register
from src.loss.types import AbstractLossFunction, BaseLossConfig

RegisteredLossFunctions: dict[
    str, (Type[AbstractLossFunction], Type[BaseLossConfig])
] = {}

register_loss_function = ft.partial(register, registry=RegisteredLossFunctions)
get_loss_function_cfgs_type = ft.partial(
    get_cfgs_type, registry=RegisteredLossFunctions
)



import os
from pathlib import Path

import equinox as eqx

from src.models import Model


def load_model(
    model: Model, model_state: eqx.nn.State, checkpoint_path: Path
) -> tuple[Model, eqx.nn.State]:
    print(f"Loading model from {checkpoint_path}")
    model = eqx.tree_deserialise_leaves(checkpoint_path, model)
    print(f"Initialized model from {checkpoint_path}")

    # Also load the corresponding state file if it exists
    if checkpoint_path.suffix == ".eqx":
        state_path = checkpoint_path.with_suffix(".eqx.state")
        if os.path.exists(state_path):
            model_state = eqx.tree_deserialise_leaves(state_path, model_state)
            print(f"Loaded model state from {state_path}")
        else:
            print(
                f"Warning: State file {state_path} not found. Using fresh model state."
            )
    else:
        print(
            "Warning: Checkpoint path doesn't end with .eqx, cannot determine state file path."
        )

    return model, model_state



from functools import partial
from typing import List

from jaxtyping import Scalar

from src.dataset.types import HeadOutput
from src.loss.types import LossFunction, BaseLossConfig


from src.loss.store import RegisteredLossFunctions


def get_loss_fns(*head_cfgs: "BaseLossConfig") -> List[LossFunction]:
    loss_fns: List[LossFunction] = []
    for head_cfg in head_cfgs:
        if len(head_cfg.loss_fns) == 0:
            raise ValueError("Must have at least one loss function for each head")

        head_loss_fns = []
        for loss_fn_cfg in head_cfg.loss_fns:
            head_loss_fns.append(
                partial(RegisteredLossFunctions[loss_fn_cfg.name][0](), loss_fn_cfg)
            )

        def head_loss_fn(y: HeadOutput, head_output: HeadOutput) -> Scalar:
            loss = 0.0
            for loss_fn in head_loss_fns:
                loss += loss_fn(y, head_output)

            return loss

        loss_fns.append(head_loss_fn)

    return loss_fns



import logging
import os
import uuid
from pathlib import Path
import wadler_lindig as wld

import hydra
import jax
import torch
from hydra.types import RunMode
from omegaconf import DictConfig
from pytorch_lightning import Trainer
from pytorch_lightning.profilers import SimpleProfiler

from src.callbacks.save_test_output import CollectTestOutputsCallback
from src.config import load_typed_root_config
from src.dataset import get_dataloaders
from src.logging import get_lightning_logger, setup_logging
from src.loss.get_loss_fns import get_loss_fns
from src.metric import get_metrics_for_heads
from src.misc.early_stopping import EarlyStoppingWithWarmup
from src.misc.load_model import load_model
from src.misc.save_checkpoint import CheckpointManager
from src.misc.step_tracker import StepTracker
from src.model_wrapper import JaxLightningWrapper
from src.models import get_model


def main(cfg_dict: DictConfig) -> float | None:
    cfg = load_typed_root_config(cfg_dict)
    setup_logging(level=cfg.log_level)
    logger = logging.getLogger(__name__)

    torch.manual_seed(cfg.seed)
    is_multirun = hydra.core.hydra_config.HydraConfig.get()["mode"] == RunMode.MULTIRUN

    logger.info(f"is_multirun: {is_multirun}")
    logger.info(jax.devices())

    if cfg.output_dir is None:
        # Set up the output directory.
        while os.path.exists(
            output_dir := (
                Path(hydra.core.hydra_config.HydraConfig.get()["runtime"]["output_dir"])
                / str(uuid.uuid4())[:4]
            )
        ):
            pass
        output_dir = output_dir
    else:
        hydra_out = str(
            hydra.core.hydra_config.HydraConfig.get()["runtime"]["output_dir"]
        ).split("outputs")
        output_dir = Path(cfg.output_dir) / Path("outputs" + hydra_out[1])

    os.makedirs(output_dir, exist_ok=True)

    cfg.output_dir = output_dir
    lightning_logger = get_lightning_logger(
        cfg=cfg, output_dir=output_dir, is_multirun=is_multirun
    )

    # For keeping track of global number of steps
    step_tracker = StepTracker()

    if cfg.use_compilation_cache:
        os.makedirs(".cache/jax_cache", exist_ok=True)
        jax.config.update("jax_compilation_cache_dir", ".cache/jax_cache")
        jax.config.update(
            "jax_persistent_cache_min_entry_size_bytes", int(1e10)
        )  # 10GB
        jax.config.update("jax_persistent_cache_min_compile_time_secs", 1.0)
        jax.config.update(
            "jax_persistent_cache_enable_xla_caches",
            "xla_gpu_per_fusion_autotune_cache_dir",
        )

    key = jax.random.PRNGKey(cfg.seed)
    key, data_key = jax.random.split(key, 2)
    train_dataloader, val_dataloader, test_dataloader = get_dataloaders(
        cfg=cfg.dataset,
        batch_size_train_val=cfg.train.batch_size,
        batch_size_test=cfg.test.batch_size,
        key=data_key,
        step_tracker=step_tracker,
        logger=lightning_logger,
    )

    key, pl_key, model_key = jax.random.split(key, 3)
    model, model_state, filter_spec, param_count = get_model(
        cfg=cfg.model,
        input_feature_dim=train_dataloader.input_feature_dim,
        input_dtype=train_dataloader.dtype,
        target_feature_dim=train_dataloader.target_feature_dim,
        key=model_key,
        timesteps=train_dataloader.input_timesteps,
        num_embeddings=train_dataloader.num_embeddings,
    )
    # add param count to wandb config
    lightning_logger.experiment.config.update({"param_count": param_count})

    if cfg.load_from_checkpoint is not None:
        model, model_state = load_model(model, model_state, cfg.load_from_checkpoint)

    pl_model = JaxLightningWrapper(
        model=model,
        model_state=model_state,
        model_filter_spec=filter_spec,
        test_config=cfg.test,
        train_config=cfg.train,
        loss_fns=get_loss_fns(*cfg.model.heads),
        step_tracker=step_tracker,
        key=pl_key,
    )

    callbacks = []
    callbacks += train_dataloader.callbacks or []

    # add callback to save test outputs
    callbacks.append(CollectTestOutputsCallback(save_dir=output_dir))

    # Set up checkpointing.
    checkpoints = CheckpointManager(
        output_dir=output_dir,
        cfg=cfg.train.checkpoint,
        step_tracker=step_tracker,
    )
    callbacks.append(checkpoints)

    if cfg.train.early_stopping is not None:
        callbacks.append(
            EarlyStoppingWithWarmup(
                monitor=cfg.train.early_stopping.metric,
                min_delta=cfg.train.early_stopping.min_delta,
                patience=cfg.train.early_stopping.patience,
                verbose=cfg.train.early_stopping.verbose,
                mode=cfg.train.early_stopping.mode,
                warmup_steps=cfg.train.early_stopping.warmup_steps,
            )
        )

    # Profiler
    profiler = SimpleProfiler(dirpath=output_dir / "profiler", filename="profiler")

    callbacks += get_metrics_for_heads(model)

    steps_per_epoch = len(train_dataloader)

    if (
        cfg.train.check_val_every_n_epoch is not None
        and cfg.train.val_check_interval > steps_per_epoch
    ):
        logger.info(f"Clamping validation frequency to {steps_per_epoch}")
        cfg.train.val_check_interval = steps_per_epoch

    trainer = Trainer(
        max_steps=cfg.train.max_steps or -1,
        max_epochs=cfg.train.max_epochs,
        min_steps=cfg.train.min_steps,
        min_epochs=cfg.train.min_epochs,
        logger=lightning_logger,
        val_check_interval=min(cfg.train.val_check_interval, steps_per_epoch),
        check_val_every_n_epoch=cfg.train.check_val_every_n_epoch,
        callbacks=callbacks,
        enable_model_summary=False,  # model summary does not work with jax
        profiler=profiler,
        log_every_n_steps=min(cfg.train.log_every_n_steps, steps_per_epoch),
    )

    logger.info(f"Config: {wld.pformat(cfg)}")

    best_val_metric = None
    if cfg.mode == "train" or cfg.mode == "all":
        trainer.fit(
            pl_model,
            train_dataloader,
            val_dataloader,
        )

        try:
            # the MAD benchmark does its sweep on the test accuracy (because the test data deviates from the validation data). It furthermore just uses the last model for evaluation.
            if not cfg.sweep_on_test:
                # load best checkpoint for testing
                model_path, state_path, best_val_metric = (
                    checkpoints.get_best_checkpoint()
                )
                pl_model.model, pl_model.model_state = load_model(
                    pl_model.model, pl_model.model_state, model_path
                )
        except Exception as e:
            logger.error(
                f"Error loading best checkpoint, Keeping most recent model loaded: {e}"
            )

    if cfg.mode == "test" or cfg.mode == "all":
        test_results = trainer.test(pl_model, test_dataloader)

    # Finish wandb experiment manually (required for doing multiruns)
    lightning_logger.experiment.finish()

    # We need to return the best metric here for the hyperparameter optimization
    if not cfg.sweep_on_test:
        return best_val_metric
    # The MAD benchmark should use the test accuracy
    else:
        return test_results[0]["test/multiclass/masked_accuracy"]


if __name__ == "__main__":
    main()



from multiprocessing import RLock

# TODO: make this work with multiple GPU training


class StepTracker:
    lock: RLock
    step: int

    def __init__(self):
        self.lock = RLock()
        self.step = 0

    def set_step(self, step: int) -> None:
        with self.lock:
            self.step = step

    def get_step(self) -> int:
        with self.lock:
            return self.step



from dataclasses import dataclass, field
from typing import List, Literal

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from src.dataset.types import UnbatchedInput, UnbatchedTarget

from .types import Head, HeadBaseConfig

from .store import register_head


@dataclass
class RegressionHeadConfig(HeadBaseConfig):
    name: Literal["regression"]
    linear_output: bool = False
    output_step: int = 1
    metrics: List[str] = field(default_factory=lambda: ["mse", "rel_mse", "l1_loss"])


@register_head(cfg=RegressionHeadConfig)
class RegressionHead(Head):
    linear_layer: eqx.nn.Linear
    # linear_layer_2: eqx.nn.Linear
    # time_layer: eqx.nn.Linear
    output_step: int
    linear_output: bool
    is_gaussian_nll: bool

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        cfg: RegressionHeadConfig,
        *,
        key: PRNGKeyArray,
        timesteps: int,
        **kwargs,
    ):
        self.name = cfg.name
        self.output_step = cfg.output_step
        self.linear_output = cfg.linear_output
        self.metrics = cfg.metrics
        self.is_gaussian_nll = "gaussian_nll" in [
            loss_fn.name for loss_fn in cfg.loss_fns
        ]
        if cfg.output_dim is not None and cfg.output_dim != input_dim:
            print(
                f"Manually overriding {cfg.name} head output_dim to {cfg.output_dim} (dataset natural output dim is {output_dim})"
            )
            output_dim = cfg.output_dim
        self.linear_layer = eqx.nn.Linear(input_dim, output_dim, key=key)
        # self.linear_layer_2 = eqx.nn.Linear(cfg.hidden_dim, output_dim, key=key)
        # self.time_layer = eqx.nn.Linear(timesteps, 1, key=key)

    def __call__(self, x: UnbatchedInput, key: PRNGKeyArray) -> UnbatchedTarget:
        # x = x[self.output_step - 1 :: self.output_step]
        # x = jax.vmap(self.linear_layer_2)(x)[-1]
        # x = jax.nn.relu(x)
        # if not self.linear_output:
        #     x = jax.nn.tanh(x)
        # x = jax.vmap(self.time_layer)(x.T).T
        x = x.mean(axis=0)
        x = self.linear_layer(x)  # (batch, 6) -> (batch, 4) # 6*4 = 24/36

        if self.is_gaussian_nll:
            # apply softplus to second half of last dimensino
            x1, x2 = jnp.split(x, 2, axis=-1)
            x2 = jax.nn.softplus(x2)
            x = jnp.concatenate([x1, x2], axis=-1)

        return x



from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping


class EarlyStoppingWithWarmup(EarlyStopping):
    """
    EarlyStopping, except don't watch the first `warmup_steps` steps.

    if warmup_steps is None, this is the same as EarlyStopping.
    """

    warmup_steps: int | None = None

    def __init__(self, warmup_steps: int | None = None, **kwargs):
        super().__init__(**kwargs)
        self.warmup_steps = warmup_steps

    def on_validation_end(self, trainer: Trainer, pl_module):
        if (
            self._check_on_train_epoch_end
            or self._should_skip_check(trainer)
            or (
                self.warmup_steps is not None
                and trainer.global_step < self.warmup_steps
            )
        ):
            return
        self._run_early_stopping_check(trainer)



from typing import Any, List

import equinox as eqx
import jax
import jax.numpy as jnp
import optax
import pytorch_lightning as pl
from jaxtyping import PRNGKeyArray

from src.config import TestCfg, TrainCfg
from src.dataset.types import (
    Batch,
    BatchedInput,
    BatchedTargets,
    ModelOutput,
)
from src.loss.types import LossFunction
from src.misc.step_tracker import StepTracker
from src.models.model import Model
from src.optimizer import get_optimizer
from src.types import JaxLightningStepOutput


class JaxLightningWrapper(pl.LightningModule):
    config: TrainCfg
    model: Model
    model_state: eqx.nn.State
    model_filter_spec: dict[str, bool]
    test_config: TestCfg
    train_config: TrainCfg
    step_tracker: StepTracker
    loss_fns: List[LossFunction]

    def __init__(
        self,
        model: Model,
        model_state: eqx.nn.State,
        model_filter_spec: dict[str, bool],
        test_config: TestCfg,
        train_config: TrainCfg,
        loss_fns: List[LossFunction],
        step_tracker: StepTracker,
        key: PRNGKeyArray,
    ):
        super().__init__()
        # we need to manually call the optimizer step using Optax instead of the lightning trainer
        self.automatic_optimization = False

        self.model = model
        self.model_state = model_state
        self.model_filter_spec = model_filter_spec
        self.test_config = test_config
        self.train_config = train_config
        self.loss_fns = loss_fns
        self.step_tracker = step_tracker

        self.key, self.train_key, self.val_key, self.test_key = jax.random.split(key, 4)

        self.configure_optimizers()

    def configure_optimizers(self):
        self.optim, self.opt_state, self.lr_schedule = get_optimizer(
            self.train_config, self.model
        )

    def training_step(self, batch: Batch) -> JaxLightningStepOutput:
        X, y = batch["input"], batch["target"]
        params = batch.get("params", None)

        # with jax.profiler.trace("scratch/jax-trace", create_perfetto_link=True):
        (
            self.model,
            self.model_state,
            self.opt_state,
            loss,
            outputs,
            grad_norms_mean,
            grad_norms_max,
        ) = JaxLightningWrapper.apply_training_step(
            model=self.model,
            model_filter_spec=self.model_filter_spec,
            model_state=self.model_state,
            X=X,
            y=y,
            loss_fns=self.loss_fns,
            key=self.train_key,
            opt_state=self.opt_state,
            opt_update=self.optim.update,
        )
        # outputs.block_until_ready()

        self.log("train/grad_norms_mean", grad_norms_mean.item(), prog_bar=True)
        self.log("train/grad_norms_max", grad_norms_max.item(), prog_bar=True)
        self.log("train/loss", loss.item(), prog_bar=True)
        lr = self.lr_schedule(self.global_step)
        if isinstance(lr, jax.Array):
            lr = lr.item()
        self.log("train/lr", lr, prog_bar=True)

        # this is needed for the lightning trainer to update the global step (https://github.com/ludwigwinkler/JaxLightning/issues/1)
        self.optimizers().step()
        self.step_tracker.set_step(self.global_step)

        return JaxLightningStepOutput(outputs=outputs, params=params)

    def validation_step(self, batch: Batch):
        X, y = batch["input"], batch["target"]
        params = batch.get("params", None)
        inference_model = eqx.tree_inference(self.model, value=True)
        diff_model, static_model = eqx.partition(
            inference_model, self.model_filter_spec
        )

        (loss, (_, outputs)), _ = eqx.filter_jit(JaxLightningWrapper.apply_loss_fn)(
            diff_model=diff_model,
            static_model=static_model,
            model_state=self.model_state,
            X=X,
            y=y,
            loss_fns=self.loss_fns,
            key=self.val_key,
        )

        self.log(
            "val/loss",
            jnp.asarray(loss).item(),
            prog_bar=True,
            batch_size=X.shape[0],
        )

        return JaxLightningStepOutput(outputs=outputs, params=params)

    def test_step(self, batch: Batch, batch_idx: int):
        X, y = batch["input"], batch["target"]
        params = batch.get("params", None)
        inference_model = eqx.tree_inference(self.model, value=True)
        # For validation, we don't update the model, just compute loss
        diff_model, static_model = eqx.partition(
            inference_model, self.model_filter_spec
        )

        (loss, (_, outputs)), _ = eqx.filter_jit(JaxLightningWrapper.apply_loss_fn)(
            diff_model=diff_model,
            static_model=static_model,
            model_state=self.model_state,
            X=X,
            y=y,
            loss_fns=self.loss_fns,
            key=self.test_key,
        )

        return JaxLightningStepOutput(outputs=outputs, params=params)

    @staticmethod
    def batched_forward(
        model: Model,
        X: BatchedInput,
        model_state: eqx.nn.State,
        key: PRNGKeyArray,
    ) -> tuple[ModelOutput, eqx.nn.State]:
        """
        Static (pure) function that vmaps over batch dimension.

        this function should be partially applied and returned
        """

        keys = jax.random.split(key, X.shape[0])
        outputs, model_state = jax.vmap(
            model,
            axis_name="batch",
            in_axes=(0, None, 0),
            out_axes=(0, None),
        )(X, model_state, keys)

        return outputs, model_state

    @staticmethod
    @eqx.filter_value_and_grad(has_aux=True)
    def apply_loss_fn(
        diff_model: Model,
        static_model: Model,
        model_state: eqx.nn.State,
        X: BatchedInput,
        y: BatchedTargets,
        loss_fns: List[LossFunction],
        key: PRNGKeyArray,
    ) -> tuple[jnp.ndarray, tuple[eqx.nn.State, ModelOutput]]:
        model = eqx.combine(diff_model, static_model)
        outputs, model_state = JaxLightningWrapper.batched_forward(
            model=model, X=X, model_state=model_state, key=key
        )

        loss = 0.0
        n_heads = len(loss_fns)
        for i, head_loss_fn in enumerate(loss_fns):
            # wrap around if there are more heads than targets
            loss += head_loss_fn(y[i % n_heads], outputs[i])

        return loss, (model_state, outputs)

    @staticmethod
    @eqx.filter_jit
    def apply_training_step(
        model: Model,
        model_filter_spec: dict[str, bool],
        model_state: Any,
        X: BatchedInput,
        y: BatchedTargets,
        loss_fns: List[LossFunction],
        key: PRNGKeyArray,
        opt_state: optax.OptState,
        opt_update: optax.TransformUpdateExtraArgsFn,
        **kwargs,
    ) -> tuple[
        Model, eqx.nn.State, optax.OptState, jnp.ndarray, ModelOutput, float, float
    ]:
        """
        Run model forward pass, compute loss and one optimizer step.

        To be autodiffed this needs to be a static (pure) function.
        """
        diff_model, static_model = eqx.partition(model, model_filter_spec)

        (loss, (new_model_state, output)), grads = JaxLightningWrapper.apply_loss_fn(
            diff_model,
            static_model=static_model,
            model_state=model_state,
            X=X,
            y=y,
            loss_fns=loss_fns,
            key=key,
        )

        grad_norms = jnp.array(
            [
                jnp.linalg.norm(x)
                for x in jax.tree.leaves(eqx.filter(grads, eqx.is_inexact_array))
            ]
        )

        grad_norms_mean = jnp.mean(grad_norms)
        grad_norms_max = jnp.max(grad_norms)

        updates, opt_state = opt_update(grads, opt_state, diff_model)
        model = eqx.apply_updates(diff_model, updates)
        model = eqx.combine(model, static_model)
        return (
            model,
            new_model_state,
            opt_state,
            loss,
            output,
            grad_norms_mean,
            grad_norms_max,
        )



from typing import Type
import functools as ft

from src.misc.stores import get_cfgs_type, register
from src.models.encoder.types import Encoder, EncoderBaseConfig

RegisteredEncoders: dict[str, (Type[Encoder], Type[EncoderBaseConfig])] = {}

register_encoder = ft.partial(register, registry=RegisteredEncoders)
get_encoder_cfgs_type = ft.partial(get_cfgs_type, registry=RegisteredEncoders)



import functools as ft

import jax
import jax.numpy as jnp
from jaxtyping import Scalar

from src.dataset.types import UnbatchedTarget
from src.metric.types import CustomMetric, Metric

RegisteredMetrics: dict[str, CustomMetric] = {}


def metric(name: str = None, prog_bar: bool = False, jit: bool = False):
    """Decorator to register a metric function."""

    def decorator(func: Metric) -> Metric:
        metric_name = name or func.__name__

        RegisteredMetrics[metric_name] = ft.partial(
            CustomMetric,
            metric=jax.jit(func) if jit else func,
            metric_name=metric_name,
            prog_bar=prog_bar,
        )

        return func

    return decorator


@metric()
def mse(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    return jnp.mean((pred - target) ** 2)


@metric()
def rel_mse(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    return jnp.mean((pred - target) ** 2) / jnp.mean(target**2)


@metric()
def l1_loss(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    return jnp.mean(jnp.abs(pred - target))


@metric()
def accuracy(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    return jnp.mean(target == pred)


# TODO: actually, the pred and target seem to be batched....
@metric()
def masked_accuracy(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    # y has shape T, y_hat has shape T x C.

    # TODO: the average should be taken over all samples in the end, not just one batch. If we do it this way, we should do a moving average.

    # there are num_tokens_to_copy * batch_size valid tokens (can check by printing jnp.sum(mask))
    mask = target != -100

    # mask out the tokens with -100
    target = target[mask]
    pred = pred[mask]

    # compute the accuracy over the valid tokens
    # accuracy = jnp.mean(target == jnp.argmax(pred, axis=-1))

    return accuracy(target, jnp.argmax(pred, axis=-1))


# TODO: this is a batched target
@metric()
def one_hot_accuracy(
    target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs
) -> Scalar:
    return accuracy(jnp.argmax(target, axis=-1), jnp.argmax(pred, axis=-1))


@metric()
def false_positive_rate(
    target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs
) -> Scalar:
    return jnp.mean(target == 0 & pred == 1)


@metric()
def false_negative_rate(
    target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs
) -> Scalar:
    return jnp.mean(target == 1 & pred == 0)


@metric()
def true_positive_rate(
    target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs
) -> Scalar:
    return jnp.mean(target == 1 & pred == 1)


@metric()
def true_negative_rate(
    target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs
) -> Scalar:
    return jnp.mean(target == 0 & pred == 0)


@metric()
def f1_score(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    tp = true_positive_rate(target, pred)
    fp = false_positive_rate(target, pred)
    fn = false_negative_rate(target, pred)
    return 2 * tp / (2 * tp + fp + fn)


@metric()
def l1_magnitude(target: UnbatchedTarget, pred: UnbatchedTarget, **kwargs) -> Scalar:
    return jnp.mean(jnp.abs(pred))



import jax
import jax.numpy as jnp
import jax.scipy as jsp


def transform_intervals(x: jnp.ndarray) -> jnp.ndarray:
    """
    Linearly transforms the interval (-1,1) on to (0,1).
    """
    return 0.5 + 0.5 * x


def inv_tril_matrix(L: jnp.ndarray) -> jnp.ndarray:
    """
    Compute the inverses of a batch of lower triangular matrices using JAX.

    Parameters:
    L : jax.numpy.ndarray
        A 3D array of shape (batch_size, n, n) with lower triangular matrices (assumed nonsingular).

    Returns:
    jax.numpy.ndarray
        The batch of inverses, shape (batch_size, n, n).
    """

    def invert_single(lower_matrix):
        n = lower_matrix.shape[-1]
        I = jnp.eye(n)
        return jsp.linalg.solve_triangular(lower_matrix, I, lower=True)

    # Vectorize over the batch dimension
    invert_batched = jax.vmap(invert_single)

    return invert_batched(L)



from typing import Type
import functools as ft

from src.misc.stores import get_cfgs_type, register
from src.dataset.types import StandardDataset, AbstractDatasetConfig

RegisteredDatasets: dict[str, (Type[StandardDataset], Type[AbstractDatasetConfig])] = {}

register_dataset = ft.partial(register, registry=RegisteredDatasets)
get_dataset_cfgs_type = ft.partial(get_cfgs_type, registry=RegisteredDatasets)



from dataclasses import dataclass, field
from typing import List, Literal

import equinox as eqx
import jax
from jaxtyping import PRNGKeyArray

from src.dataset.types import UnbatchedInput, UnbatchedTarget

from .types import Head, HeadBaseConfig

from .store import register_head


@dataclass
class MultiClassHeadConfig(HeadBaseConfig):
    name: Literal["multiclass"]
    metrics: List[str] = field(default_factory=lambda: ["masked_accuracy"])


@register_head(cfg=MultiClassHeadConfig)
class MultiClassHead(Head):
    linear_layer: eqx.nn.Linear

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        cfg: MultiClassHeadConfig,
        *,
        key: PRNGKeyArray,
        **kwargs,
    ):
        self.name = cfg.name
        self.metrics = cfg.metrics
        self.linear_layer = eqx.nn.Linear(input_dim, output_dim, key=key)

    def __call__(self, x: UnbatchedInput, key: PRNGKeyArray) -> UnbatchedTarget:
        x = jax.vmap(self.linear_layer)(x)  # shape (T, input_dim) -> (T, output_dim)

        # now x has shape T x output_dim. Apply softmax across the last dimension
        x = jax.nn.log_softmax(x, axis=-1)
        return x



from typing import Callable, List, Literal, Tuple

import jax.numpy as jnp
from pytorch_lightning import Callback, LightningModule, Trainer

from src.dataset.types import Batch, UnbatchedTarget
from src.types import JaxLightningStepOutput

Metric = Callable[[UnbatchedTarget, UnbatchedTarget], float]


class CustomMetric(Callback):
    """
    Helper class to handle metric logging for each head.
    """

    def __init__(
        self,
        heads: List[Tuple[str, int]],  # (head_name, head_index)
        metric: Metric,
        metric_name: str,
        prog_bar: bool = False,
    ):
        super().__init__()
        self.metric = metric
        self.metric_name = metric_name
        self.heads = heads
        self.prog_bar = prog_bar

    def log_metric(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: JaxLightningStepOutput,  # for each head
        batch: Batch,
        batch_idx: int,
        stage: Literal["train", "val", "test"],
    ) -> None:
        inputs, target = batch["input"], batch["target"]
        params = batch.get("params", None)

        model_outputs = outputs["outputs"]

        n_heads = len(target)
        for head_name, head_index in self.heads:
            head_output = model_outputs[head_index]
            head_target = target[head_index % n_heads]

            metric_value = self.metric(
                target=head_target,
                pred=head_output,
                params=params,
                input=inputs,
            )

            if isinstance(metric_value, jnp.ndarray):
                metric_value = metric_value.item()

            pl_module.log(
                f"{stage}/{head_name}/{self.metric_name}",
                metric_value,
                prog_bar=self.prog_bar,
                batch_size=inputs.shape[0],
            )

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self.log_metric(trainer, pl_module, outputs, batch, batch_idx, stage="train")

    def on_validation_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self.log_metric(trainer, pl_module, outputs, batch, batch_idx, stage="val")

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self.log_metric(trainer, pl_module, outputs, batch, batch_idx, stage="test")



import logging
import jax.random as jr
from jaxtyping import PRNGKeyArray
from pytorch_lightning.loggers import Logger

from src.misc.step_tracker import StepTracker

from .types import CustomDataloader, Stage
from .util import numpy_collate
from src.dataset.store import get_dataset_cfgs_type, RegisteredDatasets

from src.misc.stores import import_all_siblings

logger = logging.getLogger(__name__)


import_all_siblings(__name__, __file__)
DatasetConfig = get_dataset_cfgs_type()


def get_dataloader(
    cfg: DatasetConfig,
    batch_size: int,
    stage: Stage,
    key: PRNGKeyArray,
    step_tracker: StepTracker | None,
    logger: Logger | None = None,
) -> CustomDataloader:
    try:
        dataset = RegisteredDatasets[cfg.name][0](
            cfg=cfg, stage=stage, step_tracker=step_tracker, key=key
        )
    except KeyError:
        logger.error(f"Dataset {cfg.name} not found")
        raise ValueError(f"Dataset {cfg.name} not found")

    dataloader = CustomDataloader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=cfg.shuffle if stage == "train" else False,
        collate_fn=numpy_collate,
        input_feature_dim=dataset.input_feature_dim,
        target_feature_dim=dataset.target_feature_dim,
        input_timesteps=dataset.input_timesteps,
        dtype=dataset.dtype,
        callbacks=dataset.callbacks,
        logger=logger,
        num_embeddings=dataset.num_embeddings,
        num_workers=cfg.num_workers,
        # multiprocessing_context="spawn",  # needed for jax to work properly with multiple dataloaders
        # persistent_workers=True,
        multiprocessing_context="spawn"
        if cfg.num_workers > 0
        else None,  # needed for jax to work properly with multiple dataloaders
        persistent_workers=cfg.num_workers
        > 0,  # Only use persistent workers when num_workers > 0
    )
    return dataloader


def get_dataloaders(
    cfg: DatasetConfig,
    batch_size_train_val: int,
    batch_size_test: int | None,
    key: PRNGKeyArray,
    step_tracker: StepTracker | None,
    logger: Logger | None = None,
) -> tuple[CustomDataloader, CustomDataloader, CustomDataloader]:
    key, train_key, val_key, test_key = jr.split(key, 4)
    train_dataloader = get_dataloader(
        cfg,
        batch_size=batch_size_train_val,
        stage="train",
        key=train_key,
        step_tracker=step_tracker,
        logger=logger,
    )
    val_dataloader = get_dataloader(
        cfg,
        batch_size=batch_size_train_val,
        stage="val",
        key=val_key,
        step_tracker=step_tracker,
        logger=logger,
    )
    test_dataloader = get_dataloader(
        cfg,
        batch_size=batch_size_test or batch_size_train_val,
        stage="test",
        key=test_key,
        step_tracker=step_tracker,
        logger=logger,
    )

    return train_dataloader, val_dataloader, test_dataloader



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, Tuple, TypedDict

import jax.numpy as jnp
from jaxtyping import Array, Float, PRNGKeyArray
from pytorch_lightning import Callback
from pytorch_lightning.loggers import Logger
from torch.utils.data import DataLoader
from torch.utils.data import Dataset as TorchDataset

from src.misc.step_tracker import StepTracker

Stage = Literal["train", "val", "test"]

UnbatchedInput = Float[Array, "timesteps features"]
BatchedInput = Float[UnbatchedInput, "batch_size"]

EncoderOutput = Float[Array, "timesteps features"]
BatchedEncoderOutput = Float[EncoderOutput, "batch_size"]

UnbatchedTarget = Float[Array, "output_features"]
BatchedTarget = Float[UnbatchedTarget, "batch_size"]

UnbatchedTargets = Tuple[UnbatchedTarget, ...]
BatchedTargets = Tuple[BatchedTarget, ...]

HeadOutput = Float[Array, "batch_size output_features"]
ModelOutput = Tuple[HeadOutput, ...]

UnbatchedParams = dict[str, Float[Array, "1"]]
BatchedParams = dict[str, Float[UnbatchedParams, " batch_size"]]


@dataclass
class DatasetItem:
    input: UnbatchedInput
    target: UnbatchedTarget
    params: UnbatchedParams | None


class ShimedDatasetItem(DatasetItem):
    input: UnbatchedInput
    target: Tuple[UnbatchedTarget, ...]
    params: UnbatchedParams | None

    def __init__(
        self,
        input: UnbatchedInput,
        target: Tuple[UnbatchedTarget, ...] | UnbatchedTarget,
        params: UnbatchedParams | None = None,
    ):
        super().__init__(input, target, params)

        if isinstance(target, (tuple, list)):
            self.target = target
        else:
            self.target = (target,)


@dataclass
class AbstractDatasetConfig:
    name: str
    root_path: Path
    shuffle: bool
    num_workers: int = field(default=4, init=False)


class Batch(TypedDict):
    input: BatchedInput
    target: BatchedTargets
    params: BatchedParams | None


class StandardDataset(TorchDataset, ABC):
    @abstractmethod
    def __init__(
        self,
        cfg: AbstractDatasetConfig,
        stage: Stage,
        step_tracker: StepTracker | None,
        key: PRNGKeyArray,
    ) -> None:
        pass

    @abstractmethod
    def __getitem__(self, index: int) -> ShimedDatasetItem:
        pass

    @property
    @abstractmethod
    def input_feature_dim(self) -> int:
        """The number of features in the input data. Needs to be implemented by the dataset."""
        pass

    @property
    @abstractmethod
    def input_timesteps(self) -> int:
        """The number of timesteps in the input data. Needs to be implemented by the dataset."""
        pass

    @property
    @abstractmethod
    def target_feature_dim(self) -> int:
        """The number of classes in the output data. Needs to be implemented by the dataset."""
        pass

    @property
    @abstractmethod
    def dtype(self) -> jnp.dtype:
        """The dtype of the input data. Needs to be implemented by the dataset."""
        pass

    @property
    def callbacks(self) -> list[Callback]:
        """
        Returns a list of callbacks e.g. for visualization etc.
        """
        return []

    @property
    def num_embeddings(self) -> int | None:
        """The number of embeddings in the input data. Needs to be implemented by the dataset."""
        return None


class CustomDataloader(DataLoader):
    """
    Custom dataloader with extra properties

    - input_dim: int
    - output_dim: int
    - timesteps: int
    - dtype: jnp.dtype
    """

    input_feature_dim: int
    target_feature_dim: int
    input_timesteps: int
    dtype: jnp.dtype
    callbacks: list[Callback]
    num_embeddings: int | None

    def __init__(
        self,
        input_feature_dim: int,
        target_feature_dim: int,
        input_timesteps: int,
        dtype: jnp.dtype,
        callbacks: list[Callback],
        num_embeddings: int | None,
        logger: Logger | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.input_feature_dim = input_feature_dim
        self.target_feature_dim = target_feature_dim
        self.input_timesteps = input_timesteps
        self.dtype = dtype
        self.callbacks = callbacks
        self.logger = logger
        self.num_embeddings = num_embeddings



from dataclasses import dataclass
from typing import Literal

import equinox as eqx
from jaxtyping import PRNGKeyArray


from .store import register_encoder
from src.dataset.types import UnbatchedInput, UnbatchedTarget
from src.models.encoder.types import Encoder, EncoderBaseConfig


@dataclass
class IdentityEncoderConfig(EncoderBaseConfig):
    name: Literal["identity"]
    number: int


@register_encoder(cfg=IdentityEncoderConfig)
class IdentityEncoder(Encoder):
    _output_dim: int

    def __init__(
        self,
        input_dim: int,
        cfg: IdentityEncoderConfig,
        **kwargs,
    ):
        self._output_dim = input_dim

    @property
    def output_dim(self) -> int:
        return self._output_dim

    def __call__(
        self,
        x: UnbatchedInput,
        state: eqx.nn.State,
        key: PRNGKeyArray,
    ) -> tuple[UnbatchedTarget, eqx.nn.State]:
        return x, state



from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Tuple

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float
from matplotlib import pyplot as plt

from .types import (
    AbstractDatasetConfig,
    ShimedDatasetItem,
    Stage,
    StandardDataset,
    UnbatchedInput,
    UnbatchedTarget,
)

from src.dataset.store import register_dataset


@dataclass
class MNISTSeqConfig(AbstractDatasetConfig):
    name: Literal["mnist_seq"]
    train_val_split: float = 0.8
    timesteps: int = 128


@register_dataset(cfg=MNISTSeqConfig)
class MNISTSeq(StandardDataset):
    def __init__(self, cfg: MNISTSeqConfig, stage: Stage, **kwargs) -> None:
        self.cfg = cfg
        self.stage = stage

        # Ensure preprocessed data exists
        if (
            not (cfg.root_path / "test_data.npy").exists()
            or not (cfg.root_path / "train_data.npy").exists()
        ):
            self._preprocess()

        # Load data and labels for this stage
        self.data, self.labels = self._load_dataset(stage)

    def _load_dataset(self, stage: Stage) -> Tuple[jnp.ndarray, np.ndarray]:
        """Load data and labels for the specified stage."""
        if stage == "test":
            data = np.load(self.cfg.root_path / "test_data.npy")
            labels = np.loadtxt(self.cfg.root_path / "testlabels.txt", dtype=int)
            return data, labels

        # For train and val, we need to split the training data
        data = np.load(self.cfg.root_path / "train_data.npy")
        labels = np.loadtxt(self.cfg.root_path / "trainlabels.txt", dtype=int)

        # Split point
        split_idx = int(len(labels) * self.cfg.train_val_split)

        if stage == "train":
            return data[:split_idx], labels[:split_idx]
        elif stage == "val":
            return data[split_idx:], labels[split_idx:]
        else:
            raise ValueError(f"Invalid stage: {stage}")

    def _preprocess(self):
        """
        Read all test/train files and save as preprocessed arrays.
        Pad all sequences to the same length (128).
        """
        print("Preprocessing MNIST data...")

        # Process test data
        test_files = list(self.cfg.root_path.glob("testimg-*-inputdata.txt"))

        if len(test_files) == 0:
            raise ValueError("No files found at path: ", self.cfg.root_path)

        # Sort by the numerical index in the filename
        test_files.sort(key=lambda x: int(x.stem.split("-")[1]))

        test_data = []
        for file in test_files:
            data = np.loadtxt(file)
            data = np.pad(data, ((0, self.cfg.timesteps - data.shape[0]), (0, 0)))
            test_data.append(data)
        test_data = np.stack(test_data)
        np.save(self.cfg.root_path / "test_data.npy", test_data)

        # Process train data
        train_files = list(self.cfg.root_path.glob("trainimg-*-inputdata.txt"))
        # Sort by the numerical index in the filename
        train_files.sort(key=lambda x: int(x.stem.split("-")[1]))

        train_data = []
        for file in train_files:
            data = np.loadtxt(file)
            data = np.pad(data, ((0, self.cfg.timesteps - data.shape[0]), (0, 0)))
            train_data.append(data)
        train_data = np.stack(train_data)
        np.save(self.cfg.root_path / "train_data.npy", train_data)

        print("Done preprocessing MNIST data.")

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> ShimedDatasetItem:
        # Convert label to one-hot encoding
        onehot_labels = np.zeros(10)
        onehot_labels[self.labels[index]] = 1

        data = self.data[index]

        return ShimedDatasetItem(
            input=data,
            target=onehot_labels,
        )

    @staticmethod
    def convert_sample(
        sample: Float[Array, "sequence_length 4"],
    ) -> Float[Array, "sequence_length 4"]:
        """
        The four features are:
        - dx: change in x coordinate (implicit starting point is (0, 0))
        - dy: change in y coordinate
        - end of stroke: 1 if the stroke ends, 0 otherwise
        - end of digit: 1 if the sequence ends, 0 otherwise

        This functions converts the sample into a sequence of features.
        - x: coordinate
        - y: coordinate
        - end of stroke: 1 if the stroke ends, 0 otherwise
        - end of digit: 1 if the sequence ends, 0 otherwise

        The sequence length is the number of strokes in the sample.
        """
        x = np.cumsum(sample[:, 0])
        y = np.cumsum(sample[:, 1])
        end_of_stroke = np.concatenate([[0], np.diff(sample[:, 2])])
        end_of_digit = np.concatenate([[0], np.diff(sample[:, 3])])
        return np.stack([x, y, end_of_stroke, end_of_digit], axis=1)

    @staticmethod
    def plot(
        X: UnbatchedInput,
        y: UnbatchedTarget,
        output: UnbatchedTarget,
        path: Path | str | None = None,
    ) -> None:
        """
        Plot the first 16 samples in the batch.
        """
        # convert sequences back to images by drawing the stroke on a 28 x 28 canvas
        # arrange in 4x4 grid
        fig, axs = plt.subplots(4, 4, figsize=(10, 10))
        for i, (sample, label, output, ax) in enumerate(
            zip(
                X,
                y,
                jnp.exp(output),
                axs.flatten(),
            )
        ):
            empty_canvas = np.zeros((29, 29))
            s = MNISTSeq.convert_sample(sample)
            for x1, x2 in zip(s[:, 0], s[:, 1]):
                empty_canvas[int(x2), int(x1)] = 1
            ax.imshow(empty_canvas, cmap="gray")
            ax.set_title(f"{i}: Label {np.argmax(label)}, Pred {np.argmax(output)}")
        if path is not None:
            path = Path(path)
            path.parent.mkdir(parents=True, exist_ok=True)
            plt.savefig(path)
        else:
            plt.show()

    @property
    def target_feature_dim(self) -> int:
        return 10

    @property
    def input_feature_dim(self) -> int:
        return 4

    @property
    def dtype(self) -> jnp.dtype:
        return jnp.float32

    @property
    def input_timesteps(self) -> int:
        return self.cfg.timesteps

    @property
    def num_embeddings(self) -> int:
        return 10


if __name__ == "__main__":
    cfg = MNISTSeqConfig(
        name="mnist_seq",
        root_path=Path("data/MNIST_sequences"),
        shuffle=False,
    )
    dataset = MNISTSeq(cfg, "train")
    print(len(dataset))
    print(dataset[0]["input"].shape, dataset[0]["target"].shape)  # (46, 4) (10,)
    val_dataset = MNISTSeq(cfg, "val")
    print(len(val_dataset))
    print(dataset[0]["input"].shape, dataset[0]["target"].shape)  # (46, 4) (10,)
    dataset = MNISTSeq(cfg, "test")
    print(len(dataset))
    print(dataset[0]["input"].shape, dataset[0]["target"].shape)  # (46, 4) (10,)

    # create a batch of 16 samples and plot them
    batch = [dataset[i] for i in range(16)]
    X, y = zip(*batch)
    MNISTSeq.plot(X, y, np.zeros((16, 10)))



#!/usr/bin/env python3
"""
Command-line interface for stapax.
This script handles config path resolution for both development and installed package scenarios.
"""

import hydra
from omegaconf import DictConfig

from .config import get_config_dir
from .main import main as main_func


def cli_main():
    """CLI entry point that sets up the correct config path."""
    config_dir = get_config_dir()

    # Override the hydra decorator with the correct config path
    @hydra.main(
        version_base=None,
        config_name="main",
        config_path=str(config_dir),
    )
    def wrapped_main(cfg_dict: DictConfig) -> float | None:
        return main_func(cfg_dict)

    # Run the wrapped main function
    wrapped_main()


if __name__ == "__main__":
    cli_main()



from typing import Type
import functools as ft

from src.misc.stores import get_cfgs_type, register
from src.models.heads.types import Head, HeadBaseConfig

RegisteredHeads: dict[str, (Type[Head], Type[HeadBaseConfig])] = {}

register_head = ft.partial(register, registry=RegisteredHeads)
get_head_cfgs_type = ft.partial(get_cfgs_type, registry=RegisteredHeads)



import os
from pathlib import Path

import numpy as np
import pytorch_lightning as pl

from src.dataset.types import Batch
from src.types import JaxLightningStepOutput


class CollectTestOutputsCallback(pl.Callback):
    def __init__(self, save_dir: str):
        super().__init__()
        self.save_dir = Path(save_dir)
        self.outputs = []
        self.inputs = []
        self.targets = []

    def on_test_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: JaxLightningStepOutput,
        batch: Batch,
        batch_idx: int,
        dataloader_idx=0,
    ):
        """
        Called when the test batch ends.
        `outputs` is whatever your LightningModule's `test_step` returns.
        """

        model_outputs = outputs["outputs"]
        inputs, targets = batch["input"], batch["target"]

        self.outputs.extend(model_outputs)
        self.inputs.extend(inputs)
        self.targets.extend(targets)

    def on_test_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        """
        Called at the end of the test epoch.
        Saves all collected outputs into numpy array
        """
        self.outputs = np.array(self.outputs)
        self.inputs = np.array(self.inputs)
        self.targets = np.array(self.targets)

        os.makedirs(self.save_dir, exist_ok=True)
        np.savez(
            self.save_dir / "test_output.npz",
            outputs=self.outputs,
            inputs=self.inputs,
            targets=self.targets,
        )

        print(f"Test outputs saved to {self.save_dir}")



from dataclasses import dataclass
from typing import Callable, List, Tuple

import equinox as eqx
import jax.random as jr
from jaxtyping import PRNGKeyArray

from src.dataset.types import UnbatchedInput, UnbatchedTarget

from src.models.encoder import EncoderConfig
from src.models.encoder.types import Encoder
from src.models.heads import HeadConfig
from src.models.heads.types import Head
from src.models.sequence_mixer import SequenceMixerConfig


@dataclass
class ModelConfig:
    encoder: EncoderConfig
    heads: List[HeadConfig]
    sequence_mixer: SequenceMixerConfig | None = None


class Model(eqx.Module):
    encoder: Encoder
    heads: List[Head]

    def __init__(
        self,
        encoder: Encoder,
        heads: List[Head],
    ):
        self.encoder = encoder
        self.heads = heads

    def __call__(
        self,
        x: UnbatchedInput,
        state: eqx.nn.State,
        key: PRNGKeyArray,
    ) -> tuple[Tuple[UnbatchedTarget, ...], eqx.nn.State]:
        key, encoder_key = jr.split(key, 2)
        y, state = self.encoder(x, state=state, key=key)

        key, head_keys = jr.split(key, len(self.heads) + 1)
        outputs = []
        for head, h_key in zip(self.heads, head_keys):
            outputs.append(head(y, key=h_key))

        return tuple(outputs), state

    def filter_spec_lambda(self) -> Callable[..., bool]:
        return lambda _: True



import math
from dataclasses import dataclass
from typing import Literal

import jax
import jax.numpy as jnp
import jax.random as jr
from jax import nn, random
from jax.nn.initializers import normal
from jaxtyping import PRNGKeyArray

from .types import SequenceMixer, SequenceMixerBaseConfig
from .store import register_sequence_mixer


def simple_uniform_init(rng, shape, std=1.0):
    weights = random.uniform(rng, shape) * 2.0 * std - std
    return weights


def map_theta_to_A(thetas, G_diag, steps):
    A_plus = (
        4
        * jnp.sqrt(
            steps**4 * jnp.cos(thetas) ** (-2)
            + steps**5 * G_diag * jnp.cos(thetas) ** (-2)
        )
        - steps**2
        * (
            -4
            - 2 * steps * G_diag
            - 4 * jnp.tan(thetas) ** 2
            - 2 * steps * G_diag * jnp.tan(thetas) ** 2
        )
    ) / (2 * steps**4 * (1 + jnp.tan(thetas) ** 2))
    A_minus = (
        -4
        * jnp.sqrt(
            steps**4 * jnp.cos(thetas) ** (-2)
            + steps**5 * G_diag * jnp.cos(thetas) ** (-2)
        )
        - steps**2
        * (
            -4
            - 2 * steps * G_diag
            - 4 * jnp.tan(thetas) ** 2
            - 2 * steps * G_diag * jnp.tan(thetas) ** 2
        )
    ) / (2 * steps**4 * (1 + jnp.tan(thetas) ** 2))

    A_diag = jnp.where(thetas > jnp.pi / 2, A_plus, A_minus)

    return A_diag


# Parallel scan operations
@jax.vmap
def binary_operator(q_i, q_j):
    """Binary operator for parallel scan of linear recurrence.
    Assumes a diagonal matrix A.

    Args:
        q_i: tuple containing A_i and Bu_i at position i       (P,), (P,)
        q_j: tuple containing A_j and Bu_j at position j       (P,), (P,)
    Returns:
        new element ( A_out, Bu_out )
    """
    A_i, b_i = q_i
    A_j, b_j = q_j

    N = A_i.size // 4
    iA_ = A_i[0 * N : 1 * N]
    iB_ = A_i[1 * N : 2 * N]
    iC_ = A_i[2 * N : 3 * N]
    iD_ = A_i[3 * N : 4 * N]
    jA_ = A_j[0 * N : 1 * N]
    jB_ = A_j[1 * N : 2 * N]
    jC_ = A_j[2 * N : 3 * N]
    jD_ = A_j[3 * N : 4 * N]
    A_new = jA_ * iA_ + jB_ * iC_
    B_new = jA_ * iB_ + jB_ * iD_
    C_new = jC_ * iA_ + jD_ * iC_
    D_new = jC_ * iB_ + jD_ * iD_
    Anew = jnp.concatenate([A_new, B_new, C_new, D_new])

    b_i1 = b_i[0:N]
    b_i2 = b_i[N:]

    new_b1 = jA_ * b_i1 + jB_ * b_i2
    new_b2 = jC_ * b_i1 + jD_ * b_i2
    new_b = jnp.concatenate([new_b1, new_b2])

    return Anew, new_b + b_j


def make_linoss_im_recurrence(A_diag, step):
    r"""Compute the PxP recurrent matrix M for LinOSS-IM
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        M    (float32): the recurrent matrix (P, P)
    """
    S = 1.0 / (1.0 + step**2.0 * A_diag)
    M_11 = jnp.diag(1.0 - step**2.0 * A_diag * S)
    M_12 = jnp.diag(-1.0 * step * A_diag * S)
    M_21 = jnp.diag(step * S)
    M_22 = jnp.diag(S)

    M = jnp.block([[M_11, M_12], [M_21, M_22]])

    return M


def make_linoss_imex_recurrence(A_diag, step):
    r"""Compute the PxP recurrent matrix M for LinOSS-IMEX
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        M  (float32): the recurrent matrix (P, P)
    """
    M_11 = jnp.diag(jnp.ones_like(A_diag))
    M_12 = jnp.diag(-1.0 * step * A_diag)
    M_21 = jnp.diag(step)
    M_22 = jnp.diag(1.0 - (step**2.0) * A_diag)

    M = jnp.block([[M_11, M_12], [M_21, M_22]])

    return M


def make_damped_linoss_imex_recurrence(A_diag, G_diag, step):
    r"""Compute the PxP recurrent matrix M for Damped LinOSS-IMEX
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        G_diag  (float32):   diagonal damping matrix   (P,)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        M    (float32): the recurrent matrix (P, P)
    """
    I = jnp.ones_like(A_diag)
    S = I + step * G_diag
    M_11 = jnp.diag(1.0 / S)
    M_12 = jnp.diag(-step / S * A_diag)
    M_21 = jnp.diag(step / S)
    M_22 = jnp.diag(I - step**2 / S * A_diag)

    M = jnp.block([[M_11, M_12], [M_21, M_22]])

    return M


def apply_linoss_im(A_diag, B, input_sequence, step):
    r"""Compute the LxH output of LinOSS-IM given an LxH input.
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        B       (complex64): input matrix            (P, H)
        input_sequence (float32): input sequence of features    (L, H)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        ys (float32): the SequenceMixer states (LinOSS_IMEX layer pre-output pre-activations)      (L, P)
    """
    Bu_elements = jax.vmap(lambda u: B @ u)(input_sequence)

    schur_comp = 1.0 / (1.0 + step**2.0 * A_diag)
    M_11 = 1.0 - step**2.0 * A_diag * schur_comp
    M_12 = -1.0 * step * A_diag * schur_comp
    M_21 = step * schur_comp
    M_22 = schur_comp

    M = jnp.concatenate([M_11, M_12, M_21, M_22])

    M_elements = M * jnp.ones((input_sequence.shape[0], 4 * A_diag.shape[0]))

    F1 = M_11 * Bu_elements * step
    F2 = M_21 * Bu_elements * step
    F = jnp.hstack((F1, F2))

    _, xs = jax.lax.associative_scan(binary_operator, (M_elements, F))
    ys = xs[:, A_diag.shape[0] :]

    return ys


def apply_linoss_imex(A_diag, B, input_sequence, step):
    r"""Compute the LxH output of of LinOSS-IMEX given an LxH input.
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        B       (complex64): input matrix            (P, H)
        input_sequence (float32): input sequence of features    (L, H)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        ys (float32): the SequenceMixer states (LinOSS_IMEX layer pre-output pre-activations)      (L, P)
    """
    Bu_elements = jax.vmap(lambda u: B @ u)(input_sequence)

    A_ = jnp.ones_like(A_diag)
    B_ = -1.0 * step * A_diag
    C_ = step
    D_ = 1.0 - (step**2.0) * A_diag

    M = jnp.concatenate([A_, B_, C_, D_])

    M_elements = M * jnp.ones((input_sequence.shape[0], 4 * A_diag.shape[0]))

    F1 = Bu_elements * step
    F2 = Bu_elements * (step**2.0)
    F = jnp.hstack((F1, F2))

    _, xs = jax.lax.associative_scan(binary_operator, (M_elements, F))
    ys = xs[:, A_diag.shape[0] :]

    return ys


def apply_damped_linoss_imex(A_diag, G_diag, B, input_sequence, step):
    r"""Compute the LxH output of of Damped LinOSS-IMEX given an LxH input.
    Args:
        A_diag  (float32):   diagonal state matrix   (P,)
        G_diag  (float32):   diagonal damping matrix (P,)
        B       (complex64): input matrix            (P, H)
        input_sequence (float32): input sequence of features    (L, H)
        step    (float):     discretization time-step $\Delta_t$  (P,)
    Returns:
        ys (float32): the SequenceMixer states (LinOSS_IMEX layer pre-output pre-activations)      (L, P)
    """
    Bu_elements = jax.vmap(lambda u: B @ u)(input_sequence)

    Identity = jnp.ones_like(A_diag)
    S = Identity + step * G_diag
    M_11 = 1.0 / S
    M_12 = -step / S * A_diag
    M_21 = step / S
    M_22 = Identity - step**2 / S * A_diag

    M = jnp.concatenate([M_11, M_12, M_21, M_22])
    M_elements = M * jnp.ones((input_sequence.shape[0], 4 * A_diag.shape[0]))

    F1 = step * (1.0 / S) * Bu_elements
    F2 = step**2 * (1.0 / S) * Bu_elements
    F = jnp.hstack((F1, F2))

    _, xs = jax.lax.associative_scan(binary_operator, (M_elements, F))
    ys = xs[:, A_diag.shape[0] :]

    return ys


@dataclass
class LinOSSConfig(SequenceMixerBaseConfig):
    name: Literal["linoss"]
    dim: int
    discretization: Literal["IM", "IMEX"]
    damping: bool
    r_min: float
    theta_max: float


@register_sequence_mixer(cfg=LinOSSConfig)
class LinOSSLayer(SequenceMixer):
    A_diag: jax.Array
    G_diag: jax.Array
    B: jax.Array
    C: jax.Array
    D: jax.Array
    steps: jax.Array
    discretization: str
    damping: bool

    def __init__(
        self,
        cfg: LinOSSConfig,
        input_dim: int,  # = encoder hidden dim
        key: PRNGKeyArray,
        **kwargs,
    ):
        A_key, G_key, B_key, C_key, D_key, step_key, key = jr.split(key, 7)

        self.steps = normal(stddev=0.5)(step_key, (cfg.dim,))
        steps = nn.sigmoid(self.steps)

        if cfg.discretization == "IMEX" and cfg.damping:
            r_max = 1.0
            mags = jnp.sqrt(
                random.uniform(G_key, shape=(cfg.dim,)) * (r_max**2 - cfg.r_min**2)
                + cfg.r_min**2
            )
            self.G_diag = (1 - mags**2) / (steps * mags**2)
            G_diag = nn.relu(self.G_diag)

            theta = random.uniform(A_key, shape=(cfg.dim,)) * cfg.theta_max
            self.A_diag = map_theta_to_A(theta, G_diag, steps)
        else:
            self.G_diag = None
            self.A_diag = random.uniform(A_key, shape=(cfg.dim,))

        self.B = simple_uniform_init(
            B_key, shape=(cfg.dim, input_dim, 2), std=1.0 / math.sqrt(input_dim)
        )
        self.C = simple_uniform_init(
            C_key, shape=(input_dim, cfg.dim, 2), std=1.0 / math.sqrt(cfg.dim)
        )
        self.D = normal(stddev=1.0)(D_key, (input_dim,))

        self.discretization = cfg.discretization
        self.damping = cfg.damping

    def __call__(self, input_sequence):
        steps = nn.sigmoid(self.steps)

        B_complex = self.B[..., 0] + 1j * self.B[..., 1]
        C_complex = self.C[..., 0] + 1j * self.C[..., 1]

        if self.discretization == "IM":
            if self.damping:
                raise NotImplementedError(
                    "Discretization {} and damping = {} not implemented".format(
                        self.discretization, self.damping
                    )
                )
            else:
                A_diag = nn.relu(self.A_diag)
                ys = apply_linoss_im(A_diag, B_complex, input_sequence, steps)
        elif self.discretization == "IMEX":
            if self.damping:
                G_diag = nn.relu(self.G_diag)
                A_boundary_low = (
                    2 + steps * G_diag - 2 * jnp.sqrt(1 + steps * G_diag)
                ) / steps**2
                A_boundary_high = (
                    2 + steps * G_diag + 2 * jnp.sqrt(1 + steps * G_diag)
                ) / steps**2
                A_diag = (
                    A_boundary_low
                    + nn.relu(self.A_diag - A_boundary_low)
                    - nn.relu(self.A_diag - A_boundary_high)
                )
                ys = apply_damped_linoss_imex(
                    A_diag, G_diag, B_complex, input_sequence, steps
                )
            else:
                A_diag = nn.relu(self.A_diag)
                ys = apply_linoss_imex(A_diag, B_complex, input_sequence, steps)
        else:
            raise NotImplementedError(
                "Discretization {} not implemented".format(self.discretization)
            )

        # Apply SequenceMixer Output Operations Cx + Du
        Cy = jax.vmap(lambda x: (C_complex @ x).real)(ys)
        Du = jax.vmap(lambda u: self.D * u)(input_sequence)
        xs = Cy + Du

        return xs



from dataclasses import asdict, is_dataclass
from typing import Any, Dict, List

import numpy as np


def split(data, bounds: list):
    assert all([b < 1 for b in bounds])
    n = len(data)
    bounds = [0] + [int(n * b) for b in bounds] + [n]
    split_data = [data[bounds[i] : bounds[i + 1]] for i in range(len(bounds) - 1)]
    return tuple(split_data)


def numpy_collate(batch: List[Any]) -> Dict[str, Any]:
    """
    Collate function for PyTorch DataLoader when dataset returns a dataclass
    with fields: input, target, params.

    Args:
        batch: List of dataset items (dataclass instances).

    Returns:
        dict with keys:
        {
            "input": np.ndarray (B, T, F),
            "target": [np.ndarray(B, [T], F), ...],
            "params": { key: np.ndarray(B, [P]) }
        }
    """
    # convert dataclass -> dict
    if is_dataclass(batch[0]):
        batch = [asdict(item) for item in batch]

    inputs = np.stack([item["input"] for item in batch], axis=0)

    # targets: list of heads
    num_heads = len(batch[0]["target"])
    targets = []
    for h in range(num_heads):
        head_data = [item["target"][h] for item in batch]
        targets.append(np.stack(head_data, axis=0))

    dict_to_return = {
        "input": inputs,
        "target": targets,
    }

    # params: dict of stacked arrays
    if batch[0]["params"] is not None:
        dict_to_return["params"] = {}
        for key in batch[0]["params"].keys():
            vals = [item["params"][key] for item in batch]
            # ensure shapes align (e.g. (1,) -> (p,))
            vals = [v if v.ndim > 0 else np.expand_dims(v, 0) for v in vals]
            dict_to_return["params"][key] = np.stack(vals, axis=0)

    return dict_to_return



import logging
import uuid
from pathlib import Path
import coloredlogs
from pytorch_lightning.loggers import Logger, WandbLogger

from src.config import RootCfg
from src.logging.local_logger import LocalLogger


def setup_logging(level: int = logging.DEBUG):
    logger = logging.getLogger()
    logger.setLevel(level)

    # Avoid duplicate handlers if setup_logging() is called multiple times
    if logger.hasHandlers():
        logger.handlers.clear()

    # Console handler
    console = logging.StreamHandler()
    console.setLevel(level)

    formatter = logging.Formatter(
        "[%(asctime)s] %(levelname)-8s %(name)s: %(message)s",
        "%H:%M:%S",
    )
    console.setFormatter(formatter)
    logger.addHandler(console)

    # Optional: add coloredlogs on top of standard logging
    coloredlogs.install(level=level, logger=logger)

    return logger


def get_lightning_logger(cfg: RootCfg, output_dir: Path, is_multirun: bool) -> Logger:
    # create wandb tags
    wandb_tags = (
        [cfg.mode, cfg.model.encoder.name, cfg.dataset.name]
        + cfg.wandb.tags
        + [head.name for head in cfg.model.heads]
    )
    if hasattr(cfg.model.encoder, "sequence_mixer"):
        wandb_tags += [cfg.model.encoder.sequence_mixer.name]
    if hasattr(cfg.dataset, "task"):
        wandb_tags += [cfg.dataset.task]

    cfg.wandb.tags = wandb_tags

    # create a meaningful name for the wandb run
    wandb_name = cfg.model.encoder.name
    if hasattr(cfg.model.encoder, "sequence_mixer"):
        wandb_name += "/" + cfg.model.encoder.sequence_mixer.name
    wandb_name += "/" + cfg.dataset.name
    if hasattr(cfg.dataset, "task"):
        wandb_name += "/" + cfg.dataset.task
    if hasattr(cfg.dataset, "seq_len"):
        wandb_name += "/" + str(cfg.dataset.seq_len)

    # initialize logger
    logger = (
        WandbLogger(
            project=cfg.wandb.project,
            name=wandb_name,
            mode=cfg.wandb.mode,
            tags=cfg.wandb_tags,
            notes=cfg.wandb.notes,
            save_dir=output_dir,
            config=cfg.to_dict(),
            log_model=False,
            group=(
                "_".join(wandb_tags) + "_" + str(uuid.uuid4())[:8]
                if is_multirun
                else None
            ),
        )
        if cfg.wandb.mode != "disabled"
        else LocalLogger(config=cfg.to_dict())
    )

    return logger



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from lightning.pytorch.loggers.logger import Logger
from lightning.pytorch.utilities import rank_zero_only
from matplotlib import pyplot as plt

LOG_PATH = Path("outputs/local_logger")


@dataclass
class LocalExperiment:
    config: dict
    metrics: dict

    def finish(self):
        pass


class LocalLogger(Logger):
    def __init__(self, config: dict) -> None:
        super().__init__()

        print(f"Initializing local logger. Logging to {LOG_PATH}")

        self.experiment = LocalExperiment(config=config, metrics={})
        os.system(f"rm -r {LOG_PATH}")

    @property
    def save_dir(self):
        return str(LOG_PATH)

    @property
    def name(self):
        return "LocalLogger"

    @property
    def version(self):
        return 0

    @rank_zero_only
    def log_hyperparams(self, params):
        pass

    @rank_zero_only
    def log_metrics(self, metrics, step):
        print(f"\n\n{'=' * 10} Step {step} {'=' * 10}")
        for key, value in metrics.items():
            print(f"{step}: {key}: {value}")
        print(f"{'=' * 30}")

    @rank_zero_only
    def log_image(
        self,
        key: str,
        images: list[Any],
        step: Optional[int] = None,
        **kwargs,
    ):
        # The function signature is the same as the wandb logger's, but the step is
        # actually required.
        assert step is not None
        for index, image in enumerate(images):
            path = LOG_PATH / f"{key}/{index:0>2}_{step:0>6}.png"
            path.parent.mkdir(exist_ok=True, parents=True)
            plt.imshow(image)
            plt.savefig(path)



from dataclasses import dataclass, field
from typing import List, Literal

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from src.dataset.types import EncoderOutput, UnbatchedTarget

from .types import Head, HeadBaseConfig

from .store import register_head


@dataclass
class ClassificationHeadConfig(HeadBaseConfig):
    name: Literal["classification"]
    metrics: List[str] = field(default_factory=lambda: ["accuracy"])


@register_head(cfg=ClassificationHeadConfig)
class ClassificationHead(Head):
    linear_layer: eqx.nn.Linear
    linear_layer_2: eqx.nn.Linear

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        cfg: ClassificationHeadConfig,
        *,
        key: PRNGKeyArray,
        **kwargs,
    ):
        l1_key, l2_key = jax.random.split(key, 2)
        self.name = cfg.name
        self.metrics = cfg.metrics
        self.linear_layer = eqx.nn.Linear(input_dim, input_dim, key=l1_key)
        self.linear_layer_2 = eqx.nn.Linear(input_dim, output_dim, key=l2_key)

    def __call__(self, x: EncoderOutput, key: PRNGKeyArray) -> UnbatchedTarget:
        x = jax.vmap(self.linear_layer)(x)  # shape (features,) -> (features,)
        x = jax.nn.relu(x)
        x = jax.vmap(self.linear_layer_2)(x)  # shape (features,) -> (classes,)
        x = jnp.mean(x, axis=0)  # shape (timesteps, features) -> (features,)
        x = jax.nn.log_softmax(x, axis=-1)

        return x



from dataclasses import dataclass
from typing import Literal

import jax.numpy as jnp
from jaxtyping import Scalar

from src.dataset.types import BatchedTarget
from src.loss.store import register_loss_function
from .types import AbstractLossFunction, BaseLossConfig


@dataclass
class MSEConfig(BaseLossConfig):
    name: Literal["mse"]
    index: int | None = None


@register_loss_function(cfg=MSEConfig)
class MSE(AbstractLossFunction):
    @staticmethod
    def __call__(
        cfg: MSEConfig,
        true_y: BatchedTarget,
        pred_y: BatchedTarget,
    ) -> Scalar:
        if cfg.index is not None:
            true_y = true_y[..., cfg.index]
            pred_y = pred_y[..., cfg.index]
        return cfg.weight * jnp.mean((true_y - pred_y) ** 2)



from typing import Tuple
import logging

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from src.models.encoder import get_encoder
from src.models.heads import get_heads
from src.models.model import Model, ModelConfig
from src.models.sequence_mixer import get_sequence_mixer_cls


logger = logging.getLogger(__name__)


def get_model(
    cfg: ModelConfig,
    input_feature_dim: int,
    target_feature_dim: int,
    input_dtype: jnp.dtype,
    key: PRNGKeyArray,
    num_embeddings: int | None = None,
    **kwargs,
) -> Tuple[Model, eqx.nn.State, dict[str, bool], int]:
    """ """
    filter_spec = {}
    if (
        hasattr(cfg.encoder, "sequence_mixer")
        and cfg.encoder.sequence_mixer is not None
    ):
        sequence_mixer_cls, sequence_mixer_filter_spec = get_sequence_mixer_cls(
            cfg=cfg.encoder.sequence_mixer,
            # input_dim=cfg.encoder.hidden_dim,
            # key=key,
            # input_dtype=input_dtype,
            # **kwargs,
        )
    else:
        sequence_mixer_cls = None
        filter_spec["sequence_mixer"] = False

    encoder, encoder_filter_spec = get_encoder(
        cfg=cfg.encoder,
        key=key,
        input_dim=input_feature_dim,
        input_dtype=input_dtype,
        sequence_mixer_cls=sequence_mixer_cls,
        num_embeddings=num_embeddings,
        **kwargs,
    )

    heads = get_heads(
        *cfg.heads,
        key=key,
        input_dim=encoder.output_dim,
        output_dim=target_feature_dim,
        **kwargs,
    )

    model = Model(encoder, heads)
    filter_spec = jax.tree_util.tree_map(model.filter_spec_lambda(), model)

    param_count = sum(
        x.size for x in jax.tree.leaves(eqx.filter(model, eqx.is_inexact_array))
    )

    logger.info(f"Created model with param count: {param_count}")

    return model, eqx.nn.State(model), filter_spec, param_count



from dataclasses import dataclass
from typing import Literal

import jax.numpy as jnp
from jaxtyping import Scalar

from src.dataset.types import BatchedTarget
from src.loss.store import register_loss_function

from .types import AbstractLossFunction, BaseLossConfig


@dataclass
class CrossEntropyConfig(BaseLossConfig):
    name: Literal["cross_entropy"]
    for_each_timestep: bool = False


@register_loss_function(cfg=CrossEntropyConfig)
class CrossEntropy(AbstractLossFunction):
    @staticmethod
    def __call__(
        cfg: CrossEntropyConfig,
        true_y: BatchedTarget,
        pred_y: BatchedTarget,
    ) -> Scalar:
        # if true_y is one-hot encoded, we need to convert it back
        if true_y.ndim > 1 and not cfg.for_each_timestep:
            indices = jnp.argmax(true_y, axis=-1)
        else:
            # only batch dimension
            indices = true_y

        # Create a mask to ignore tokens marked with -100
        # -100 is commonly used as a padding/ignore token in NLP tasks
        # TODO: debug and print to check this is correct. I was not able to do this for jax reasons... @BEN, could you varify that jnp.sum(mask) = batch_size * num_tokens_to_copy?
        mask = indices != -100

        pred_y_at_target_index = jnp.take_along_axis(
            pred_y, jnp.expand_dims(indices, -1), axis=-1
        )

        # Apply the mask to exclude -100 tokens from loss computation
        # need to do the awkward jnp.where for jax reasons...
        masked_pred_y = jnp.where(mask[..., None], pred_y_at_target_index, 0.0)

        # divide by the valid tokens to get the mean loss
        return cfg.weight * -jnp.sum(masked_pred_y) / jnp.sum(mask)



from dataclasses import asdict, dataclass
from typing import Literal

import optax


@dataclass
class CosineOneCycleWarmupCfg:
    name: Literal["warmup_cosine_decay_schedule"]
    init_value: float
    peak_value: float
    warmup_steps: int
    decay_steps: int


@dataclass
class CosineOneCycleCfg:
    name: Literal["cosine_decay_schedule"]
    init_value: float
    alpha: float
    decay_steps: int


@dataclass
class ConstantCfg:
    name: Literal["constant"]
    value: float


@dataclass
class ExponentialDecayCfg:
    name: Literal["exponential_decay"]
    init_value: float
    transition_steps: int
    decay_rate: float


LR_SCHEDULES = {
    "warmup_cosine_decay_schedule": optax.schedules.warmup_cosine_decay_schedule,
    "constant": optax.constant_schedule,
    "exponential_decay": optax.exponential_decay,
    "cosine_decay_schedule": optax.schedules.cosine_decay_schedule,
}

LRScheduleCfgs = (
    CosineOneCycleWarmupCfg
    | ConstantCfg
    | CosineOneCycleCfg
    | ExponentialDecayCfg
    | int
    | float
)


def get_learning_rate_schedule(cfg: LRScheduleCfgs) -> optax.Schedule:
    if isinstance(cfg, (int, float)):
        return LR_SCHEDULES["constant"](value=cfg)
    cfg_dict = asdict(cfg)
    name = cfg_dict.pop("name")
    return LR_SCHEDULES[name](**cfg_dict)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List

import equinox as eqx
from jaxtyping import PRNGKeyArray

from src.dataset.types import (
    EncoderOutput,
    UnbatchedTarget,
)
from src.loss import LossConfig


@dataclass
class HeadBaseConfig:
    name: str
    loss_fns: List[LossConfig]
    output_dim: int | None = None
    metrics: List[str] = field(default_factory=lambda: [])


class Head(eqx.Module, ABC):
    name: str
    metrics: List[str]

    @abstractmethod
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        cfg: HeadBaseConfig,
        *,
        key: PRNGKeyArray,
        **kwargs,
    ):
        pass

    @abstractmethod
    def __call__(self, x: EncoderOutput, key: PRNGKeyArray) -> UnbatchedTarget:
        pass



import equinox as eqx
import optax

from src.config import TrainCfg
from src.misc.learning_rate import get_learning_rate_schedule
from src.models.model import Model


def get_optimizer(config: TrainCfg, model: Model):
    lr_schedule = get_learning_rate_schedule(config.lr)
    optimizer = None

    if config.weight_decay is not None:
        optimizer = optax.adamw(
            lr_schedule,
            weight_decay=config.weight_decay,
            b1=0.9,
            b2=0.98,
        )
    else:
        optimizer = optax.adam(lr_schedule, b1=0.9, b2=0.98)

    if config.clip_grad_global_norm is not None:
        optimizer = optax.chain(
            optax.clip_by_global_norm(config.clip_grad_global_norm),
            optimizer,
        )
    opt_state = optimizer.init(eqx.filter(model, eqx.is_inexact_array))

    return optimizer, opt_state, lr_schedule



from typing import Type
import functools as ft

from src.misc.stores import get_cfgs_type, register
from src.models.sequence_mixer.types import SequenceMixer, SequenceMixerBaseConfig

RegisteredSequenceMixers: dict[
    str, (Type[SequenceMixer], Type[SequenceMixerBaseConfig])
] = {}


register_sequence_mixer = ft.partial(register, registry=RegisteredSequenceMixers)
get_sequence_mixer_cfgs_type = ft.partial(
    get_cfgs_type, registry=RegisteredSequenceMixers
)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

import equinox as eqx
from pytorch_lightning import Callback, LightningModule
from pytorch_lightning.trainer import Trainer

from src.misc.step_tracker import StepTracker


@dataclass
class CheckpointCfg:
    metric: str | None = None
    mode: Literal["min", "max"] = "max"
    metric_top_k: int = 20
    step_top_k: int = 10
    every_n_steps: int | None = None
    warmup_steps: int | None = 1


class CheckpointManager(Callback):
    def __init__(self, output_dir: Path, cfg: CheckpointCfg, step_tracker: StepTracker):
        self.output_dir = output_dir / "checkpoints"
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.cfg = cfg
        self.step_tracker = step_tracker

        # Separate tracking for metric-based and step-based checkpoints
        self.metric_checkpoints: dict[float, tuple[Path, Path]] = {}
        self.step_checkpoints: dict[int, tuple[Path, Path]] = {}
        self.best_metric: float | None = None

        # Save every 1000 steps if no metric is provided.
        if cfg.metric is None and cfg.every_n_steps is None:
            self.cfg.every_n_steps = 1000

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        if self._should_skip_warmup(trainer):
            return

        # Handle metric-based saving
        if self.cfg.metric is not None:
            self._save_by_metric(trainer)

        # Handle step-based saving
        if self.cfg.every_n_steps is not None:
            self._save_by_steps(trainer)

    def _should_skip_warmup(self, trainer: Trainer) -> bool:
        current_step = self.step_tracker.get_step()
        if self.cfg.warmup_steps is not None and current_step < self.cfg.warmup_steps:
            print(f"Skipping checkpoint saving during warmup: step {current_step}")
            return True
        return False

    def _save_by_metric(self, trainer: Trainer):
        try:
            score = trainer.logged_metrics[self.cfg.metric].item()
            if self._is_better_metric(score):
                self.best_metric = score
                self._save_checkpoint(trainer, score=score, checkpoint_type="metric")
        except KeyError:
            print(
                f"Metric '{self.cfg.metric}' not found in logged metrics, saving every {self.cfg.every_n_steps or '1000'} steps"
            )
            if self.cfg.every_n_steps is None:
                self.cfg.every_n_steps = 1000

    def _save_by_steps(self, trainer: Trainer):
        current_step = self.step_tracker.get_step()
        if current_step % self.cfg.every_n_steps == 0:
            self._save_checkpoint(trainer, checkpoint_type="step")

    def _is_better_metric(self, score: float) -> bool:
        if self.best_metric is None:
            return True
        return (
            score > self.best_metric
            if self.cfg.mode == "max"
            else score < self.best_metric
        )

    def _save_checkpoint(
        self,
        trainer: Trainer,
        score: float | None = None,
        checkpoint_type: str = "step",
    ):
        self._cleanup_old_checkpoints(checkpoint_type)

        model_path, state_path = self._get_checkpoint_paths(trainer, score)

        # Save model and state
        eqx.tree_serialise_leaves(model_path, trainer.lightning_module.model)
        eqx.tree_serialise_leaves(state_path, trainer.lightning_module.model_state)

        current_step = self.step_tracker.get_step()
        # Track checkpoint
        if checkpoint_type == "metric" and score is not None:
            self.metric_checkpoints[score] = (model_path, state_path)
            print(f"Saved metric checkpoint (score={score}) at step {current_step}")
        else:
            self.step_checkpoints[current_step] = (model_path, state_path)
            print(f"Saved step checkpoint at step {current_step}")

    def _get_checkpoint_paths(
        self, trainer: Trainer, score: float | None = None
    ) -> tuple[Path, Path]:
        suffix = f"_metric_{score}" if score is not None else ""
        current_step = self.step_tracker.get_step()
        base_name = f"checkpoint_step_{current_step}{suffix}"
        return (
            self.output_dir / f"{base_name}.eqx",
            self.output_dir / f"{base_name}.eqx.state",
        )

    def _cleanup_old_checkpoints(self, checkpoint_type: str):
        if checkpoint_type == "metric":
            self._remove_excess_checkpoints(
                self.metric_checkpoints, self.cfg.metric_top_k, use_mode=True
            )
        else:
            self._remove_excess_checkpoints(
                self.step_checkpoints, self.cfg.step_top_k, use_mode=False
            )

    def _remove_excess_checkpoints(
        self, checkpoint_dict: dict, top_k: int, use_mode: bool
    ):
        if len(checkpoint_dict) >= top_k:
            # Sort checkpoints - for metrics use mode, for steps use ascending order
            reverse_sort = self.cfg.mode == "max" if use_mode else False
            sorted_items = sorted(
                checkpoint_dict.items(), key=lambda x: x[0], reverse=reverse_sort
            )

            # Remove the worst/oldest checkpoint
            key_to_remove, (model_path, state_path) = sorted_items[-1]
            self._delete_checkpoint_files(model_path, state_path)
            del checkpoint_dict[key_to_remove]

    def _delete_checkpoint_files(self, model_path: Path, state_path: Path):
        for path in [model_path, state_path]:
            if path.exists():
                os.remove(path)

    def get_best_checkpoint(self) -> tuple[Path, Path, float]:
        """Returns (model_path, state_path, best_metric)"""
        if self.best_metric is None or self.best_metric not in self.metric_checkpoints:
            raise ValueError("No best metric checkpoint available")
        model_path, state_path = self.metric_checkpoints[self.best_metric]
        return model_path, state_path, self.best_metric



from dataclasses import asdict, dataclass, field
import logging
from pathlib import Path
from typing import Any, Literal, Type, TypeVar
import importlib.resources

from dacite import Config, from_dict
from omegaconf import DictConfig, OmegaConf

from src.dataset import DatasetConfig
from src.misc.learning_rate import LRScheduleCfgs
from src.misc.save_checkpoint import CheckpointCfg

from .models import ModelConfig

OmegaConf.register_new_resolver("multiply", lambda x, y: x * y)
OmegaConf.register_new_resolver("divide", lambda x, y: x / y)
OmegaConf.register_new_resolver(
    "total_steps",
    lambda num_examples, batch_size, epochs: (int(num_examples) // int(batch_size))
    * int(epochs),
)


def get_config_dir() -> Path:
    """Get the path to the config directory in the installed package."""
    # Try relative path from this file (development)
    rel_config_path = Path(__file__).parent.parent / "configs"
    if rel_config_path.exists():
        return rel_config_path

    # Try to get the config directory from the package
    try:
        with importlib.resources.path("stapax", "config") as config_path:
            return Path(config_path)
    except Exception:
        # Final fallback
        return Path("configs")


@dataclass
class TestCfg:
    output_path: Path = Path("results")
    batch_size: int | None = None


@dataclass
class WandbCfg:
    name: str
    project: str
    mode: Literal["online", "offline", "shared", "disabled", "dryrun", "run"] = (
        "offline"
    )
    tags: list[str] = field(default_factory=list)
    notes: str | None = None


@dataclass
class EarlyStoppingCfg:
    metric: str = "val/mse"
    min_delta: float = 0.0
    patience: int = 20
    verbose: bool = True
    mode: Literal["min", "max"] = "min"
    warmup_steps: int | None = None


@dataclass
class TrainCfg:
    lr: LRScheduleCfgs
    batch_size: int
    val_check_interval: float = 1.0
    check_val_every_n_epoch: int | None = 1
    log_every_n_steps: int = 25
    clip_grad_global_norm: float | None = None
    weight_decay: float | None = None
    checkpoint: CheckpointCfg = field(default_factory=CheckpointCfg)
    early_stopping: EarlyStoppingCfg | None = None
    max_steps: int = -1  # -1 is no limit
    max_epochs: int = -1  # -1 is no limit
    min_steps: int | None = None
    min_epochs: int | None = None

    def __post_init__(self):
        # Validate configuration. Note that -1 is used in PyTorch Lightningto indicate that the max/min steps/epochs are not set.
        if self.max_steps != -1 and self.max_epochs != -1:
            raise ValueError(
                "Both max_steps and max_epochs are defined in the config. Please specify only one."
            )
        if self.min_steps is not None and self.min_epochs is not None:
            raise ValueError(
                "Both min_steps and min_epochs are defined in the config. Please specify only one."
            )


@dataclass
class RootCfg:
    wandb: WandbCfg
    mode: Literal["train", "test", "all"]
    model: ModelConfig
    train: TrainCfg
    test: TestCfg
    dataset: DatasetConfig
    seed: int = 42
    load_from_checkpoint: Path | None = None
    output_dir: Path | None = None
    use_compilation_cache: bool = True
    sweep_on_test: bool = False
    log_level: int = logging.INFO

    def to_dict(self) -> dict:
        as_dict = asdict(self)

        def _stringify(value: Any) -> Any:
            if isinstance(value, Path):
                return str(value)
            elif isinstance(value, dict):
                return {k: _stringify(v) for k, v in value.items()}
            elif isinstance(value, list):
                return [_stringify(v) for v in value]
            elif isinstance(value, tuple):
                return tuple(_stringify(v) for v in value)
            elif isinstance(value, object):
                return str(value)
            return value

        return _stringify(as_dict)


TYPE_HOOKS = {
    Path: Path,
}


T = TypeVar("T")


def load_typed_config(
    cfg: DictConfig,
    data_class: Type[T],
    extra_type_hooks: dict = {},
) -> T:
    return from_dict(
        data_class,
        OmegaConf.to_container(cfg, resolve=True),
        config=Config(type_hooks={**TYPE_HOOKS, **extra_type_hooks}),
    )


def load_typed_root_config(cfg: DictConfig, configType: Type[T] = RootCfg) -> T:
    return load_typed_config(
        cfg,
        configType,
        {},
    )



import importlib
import pkgutil
import pathlib
from typing import Any, Callable, Dict, Type, TypeVar, Union, get_args, Tuple

M = TypeVar("M")  # Model type
C = TypeVar("C")  # Config type


Registry = Dict[str, Tuple[Type[M], Type[C]]]


def import_all_siblings(pkg: str, file: str):
    pkg_path = pathlib.Path(file).parent
    for _, modname, _ in pkgutil.iter_modules([str(pkg_path)]):
        importlib.import_module(f"{pkg}.{modname}")


def register(registry: Registry[M, C], cfg: Type[C]) -> Callable[[Type[M]], Type[M]]:
    """
    Decorator that registers a model class with its config into the given registry.
    Assumes the config has a field `name: Literal["..."]`.
    """

    def decorator(model_cls: Type[M]) -> Type[M]:
        # extract name from cfg annotation
        names = get_args(cfg.__annotations__.get("name"))
        for name in names:
            registry[name] = (model_cls, cfg)
        return model_cls

    return decorator


def get_cfgs_type(registry: Registry[M, C]) -> Type[Union[C, Any]]:
    """
    Returns a Union of all registered config types.
    """
    cfg_types = [cfg for _, cfg in registry.values()]
    if not cfg_types:
        raise ValueError("No configs registered")
    return Union[tuple(cfg_types)]



from typing import List, Tuple

from lightning import Callback

from src.metric.metrics import RegisteredMetrics
from src.models.model import Model

__all__ = ["RegisteredMetrics"]


def get_metrics_for_heads(model: Model) -> list[Callback]:
    callbacks = []
    names = [head.name for head in model.heads]

    # if there are multiple heads of the same type add _1, _2, etc.
    for name in names:
        if names.count(name) > 1:
            p = 0
            for i in range(names.count(name)):
                p = names.index(name, p)
                names[p] = f"{name}_{i}"

    # add metrics for each head
    metrics_dict: dict[str, List[Tuple[str, int]]] = {}
    for i, (head, name) in enumerate(zip(model.heads, names)):
        metrics = head.metrics
        for metric in metrics:
            if metric not in metrics_dict:
                metrics_dict[metric] = []
            metrics_dict[metric].append((name, i))

    for metric_name, heads in metrics_dict.items():
        callbacks.append(RegisteredMetrics[metric_name](heads))
    return callbacks



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable

import equinox as eqx
from jaxtyping import PRNGKeyArray

from src.dataset.types import UnbatchedInput


@dataclass
class SequenceMixerBaseConfig:
    name: str


class SequenceMixer(eqx.Module, ABC):
    @abstractmethod
    def __init__(
        self,
        cfg: SequenceMixerBaseConfig,
        input_dim: int,
        key: PRNGKeyArray,
        **kwargs,
    ):
        pass

    def filter_spec_lambda(self) -> Callable[..., bool]:
        return lambda _: True

    @abstractmethod
    def __call__(self, input_sequence: UnbatchedInput) -> UnbatchedInput:
        pass
